% UC: 21180 - Computação Numérica
% e-fólio B 2024-25
%
% Aluno: 1100477 - Sergio Rosa
%
% -------------------------------------------------------------
% Função: sol_lu_mat(A, B, tol)
%
% Descrição:
%   Resolve o sistema matricial AX = B (coluna a coluna),
%   através da decomposição LU de A (sem pivotagem).
%   A = LU => Ly = b => Ux = y.
%
% Parâmetros de entrada:
%   A   - Matriz de coeficientes (nxn)
%   B   - Matriz de termos independentes (nxm)
%   tol - Tolerância para detetar singularidade de A
%
% Parâmetro de saída:
%   X   - Matriz solução (nxm) ou [] se A for considerada singular
%
% -------------------------------------------------------------
function X = sol_lu_mat(A, B, tol)
    X=[];
    % Verificação dos argumentos de entrada
    if (nargin < 3)
        error('Número de argumentos insuficiente. Usar: sol_lu_mat(A,B,tol)');
        return;
    end

    % Dimensões de A e B
    [nA, mA] = size(A);
    [nB, mB] = size(B);

    % Verificar conformidade para multiplicação AX = B
    if nA ~= mA
        error('A não é quadrada.');
        return;
    end

    % Verificar conformidade para multiplicação AX = B
    if nA ~= nB
        error('Dimensões incompatíveis: número de linhas de A e B não coincide.');
        return;
    end


    % ---------------------------------------------------------
    % 1) Testa se A é singular (com a tolerância dada)
    %    Critério: se |det(A)| < tol OU se algum pivô ~ 0,
    %    devolve [] e emite mensagem de erro.
    %    Aqui, faremos a verificação dos pivôs diretamente
    %    na construção de L e U.
    % ---------------------------------------------------------

    % Inicializa L e U
    L = eye(nA);
    U = zeros(nA);

    for k = 1:nA
        % Calcular elementos de U na linha k, à direita de k
        U(k, k:end) = A(k, k:end) - L(k,1:k-1)*U(1:k-1,k:end);

        % Teste do pivô (U(k,k)) antes de prosseguir
        if (abs(U(k,k)) < tol)
            warning('Matriz A considerada singular dentro da tolerância especificada.');
            return;
        end

        % Calcular elementos de L na coluna k, abaixo de k
        if (k < nA)
            L(k+1:end, k) = (A(k+1:end, k) - L(k+1:end,1:k-1)*U(1:k-1, k)) / U(k,k);
        end
    end

    % ---------------------------------------------------------
    % 2) Possuímos L e U bem definidos.
    %    Agora, para cada coluna j de B, resolvemos Ly = b_j
    %    e depois Ux = y.
    % ---------------------------------------------------------
    % B pode ter várias colunas (mB). Vamos resolver tudo de modo vetorial,
    % mas a lógica é:
    %   Y = L \ B  (Forward-substitution)
    %   X = U \ Y  (Backward-substitution)
    %
    % O "problema" é que não podemos usar diretamente "\" (pois é
    % justamente o que estamos implementando). Então o forward e backward
    % substitutions serão feitos manualmente, porém em formato matricial.
    % ---------------------------------------------------------

    % Step 1: Resolução Ly = B
    %   L é triangular inferior, com diag(L) = 1
    %   y = inv(L) * B, mas faremos via forward-substitution.
    Y = zeros(nA, mB);

    for j = 1:mB
        % Para cada coluna b_j de B, resolvemos L*y_j = b_j
        % via forward-substitution:
        % Inicializar y com zeros e resolver a primeira linha
        y_col = zeros(nA, 1);
        b_col = B(:, j);

        % y_1 = b_1 / L(1,1)
        y_col(1) = b_col(1);
        for i = 2:nA
            % y_i = b_i - sum_{p=1}^{i-1} L(i,p)*y_p
            y_col(i) = b_col(i) - L(i,1:i-1)*y_col(1:i-1);
        end

        Y(:, j) = y_col;
    end

    % Step 2: Resolução Ux = Y
    %   U é triangular superior
    %   x = inv(U) * y, mas faremos via backward-substitution.
    X = zeros(nA, mB);

    for j = 1:mB
        % Para cada coluna y_j, resolvemos U*x_j = y_j
        % via backward-substitution:
        % Inicializar x com zeros e resolver a última linha
        x_col = zeros(nA, 1);
        y_col = Y(:, j);

        % x_n = y_n / U(n,n)
        x_col(nA) = y_col(nA) / U(nA, nA);
        for i = nA-1 : -1 : 1
            % x_i = (y_i - sum_{p=i+1}^{n} U(i,p)*x_p) / U(i,i)
            soma = U(i, i+1:end)*x_col(i+1:end);
            x_col(i) = (y_col(i) - soma) / U(i,i);
        end

        % Guardar a solução na matriz X
        X(:, j) = x_col;
    end

end

